@@ -1,7 +1,7 @@
 from .FBRef import FBRef
 from .FiveThirtyEight import FiveThirtyEight
 from .Understat import Understat
-from .SofaScore import SofaScore
+# from .SofaScore import SofaScore
 from .ClubElo import ClubElo
 from .Capology import Capology
 from .Transfermarkt import Transfermarkt
